{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE TypeSynonymInstances #-}

module Typing where

import Poly

import Unbound.LocallyNameless hiding (Subst, apply, compose)

import Control.Monad.Trans.Class

import Data.Maybe
import Data.List (nub)
import qualified Data.Set as Set

type Context = [(ExpName, TyScheme)]

type Subst = [(TyName, Ty)]

freshfor :: (Alpha t, Rep a) => Name a -> t -> Bool
freshfor x t = x `Set.notMember` fv t

compose :: Subst -> Subst -> Subst
compose s1 s2 =  [(a, substs s1 t) | (a, t) <- s2]
              ++ [(a, t) | (a, t) <- s1, isNothing $ lookup a s2]

mgu :: Ty -> Ty -> Maybe Subst
mgu ty1 ty2 = unify [(ty1,ty2)] []
  where unify :: [(Ty,Ty)] -> Subst -> Maybe Subst
        unify [] s = Just s
        unify (constraint:rest) s =
          case constraint of
            (TyVar a, t)
              | a `freshfor` t ->
                  unify [(subst a t t1, subst a t t2) | (t1,t2) <- rest] ((a,t):s)
              | TyVar a == t ->
                  unify rest s
            (t, TyVar a) ->
              unify ((TyVar a, t) : rest) s
            (TyInt, TyInt) ->
              unify rest s
            (TyUnit, TyUnit) ->
              unify rest s
            (TyFun t1 t2, TyFun t1' t2') ->
              unify ((t1,t1') : (t2,t2') : rest) s
            (TyRef ty, TyRef ty') ->
              unify ((ty,ty') : rest) s
            _ -> Nothing

infer :: Context -> Exp -> Maybe TyScheme
infer context = runFreshMT . fmap (generalize context) . ti context
  where ti ctx (Var x) =
          case lookup x ctx of
            Nothing -> lift Nothing
            Just (TyScheme tsc) -> do
              (_,ty) <- unbind tsc
              return ([], ty)
        ti ctx (Lam bnd) = do
          (x, body) <- unbind bnd
          freshTyVar <- fresh (s2n "a" :: TyName)
          (s, bodyTy) <- ti ((x, TyScheme (bind [] (TyVar freshTyVar))):ctx) body
          return (s, TyFun (substs s (TyVar freshTyVar)) bodyTy)
        ti ctx (App e1 e2) = do
          a <- fresh (s2n "a" :: TyName)
          (s1, t1) <- ti ctx e1
          (s2, t2) <- ti (substs s1 ctx) e2
          s3 <- lift $ mgu (TyFun t2 (TyVar a)) (substs s2 t1)
          return (compose s3 (compose s2 s1), substs s3 (TyVar a))
        ti ctx (Let bnd) = do
          ((x,ee1), e2) <- unbind bnd
          let e1 = unembed ee1
          (s1, tye1) <- ti ctx e1
          let ctx' = substs s1 ctx
          let vars = nub $ filter (`freshfor` ctx') (fv tye1)
          (s2, tye2) <- ti ((x, TyScheme (bind vars tye1)):ctx') e2
          return (compose s2 s1, substs s2 tye2)
        ti ctx (Seq e1 e2) = do
          (s1, tye1) <- ti ctx e1
          s1' <- lift $ mgu tye1 TyUnit
          let s2 = compose s1' s1
          (s3, tye2) <- ti (substs s2 ctx) e2
          return (compose s3 s2, substs (compose s3 s2) tye2)
        ti ctx (Ref e) = do
          (s, ty) <- ti ctx e
          return (s, substs s (TyRef ty))
        ti ctx (Deref e) = do
          (s, ty) <- ti ctx e
          freshTyVar <- fresh (s2n "a" :: TyName)
          s' <- lift $ mgu (TyRef (TyVar freshTyVar)) ty
          return (compose s' s, substs s' (TyVar freshTyVar))
        ti ctx (Assign e1 e2) = do
          (s, tye1) <- ti ctx e1
          freshTyVar <- fresh (s2n "a" :: TyName)
          s1 <- lift $ mgu (TyRef (TyVar freshTyVar)) tye1
          let s2 = compose s1 s
          (s3, tye2) <- ti (substs s2 ctx) e2
          s4 <- lift $ mgu (substs s1 (TyVar freshTyVar)) tye2
          return (compose s4 (compose s3 s2), TyUnit)
        ti _ Unit = return ([], TyUnit)
        ti _ (Const _) = return ([], TyInt)
        ti ctx (Succ e) = do
          (s1, ty) <- ti ctx e
          s2 <- lift $ mgu ty TyInt
          return (compose s2 s1, TyInt)
        ti _ (StoreLoc _) = lift Nothing

generalize :: Context -> (Subst, Ty) -> TyScheme
generalize ctx (s,ty) =
  let ctx' = substs s ctx
      vars = nub $ filter (`freshfor` ctx') (fv ty)
  in TyScheme (bind vars ty)

wellTyped :: Exp -> Bool
wellTyped = wellTypedWith []

wellTypedWith :: Context -> Exp -> Bool
wellTypedWith ctx = isJust . infer ctx
